program Sweeper;
{$M 8192,0,655360}
{$I-,R-,Q-,S-}
uses
   Objects, DOS,
   Dirs;

const
   VStr =          '1.1';
   CfgFile =       'SWEEPER.CFG';

type
   TParsed =       Array [0..10] of Char;

   TMaskInfo =
   record
      Mask         :TParsed;
      Info         :Integer;
   end;

   TFileInfo =
   record
      Name         :PString;
      Info         :Integer;
   end;

   TTypeInfo =
   record
      Id           :String[15];
      Name         :PString;
      Count        :Integer;
      Size         :LongInt;
   end;

   TDeleteList =
   object(TCollection)
      Types        :TCollection;

      constructor Init;
   end;

   TSweeper =
   object(TProcessFiles)
      DeleteList   :TDeleteList;

      TempList     :TCollection;
      TempSize     :LongInt;

      ClusterSize  :Word;

      constructor Init;

      procedure Run;
      procedure ProcessFile; virtual;

      procedure Delete;
   end;

function  DriveValid(Drive             :Char)    :Boolean; near; assembler;
asm
   MOV  AH,19H          { Save the current drive in BL }
   INT  21H
   MOV  BL,AL
   MOV  DL,Drive        { Select the given drive }
   SUB  DL,'A'
   MOV  AH,0EH
   INT  21H
   MOV  AH,19H          { Retrieve what DOS thinks is current }
   INT  21H
   xor  cx,cx           { Assume false }
   CMP  AL,DL           { Is the current drive the given drive? }
   JNE  @@1
   inc  cx              { It is, so the drive is valid }
   MOV  DL,BL           { Restore the old drive }
   MOV  AH,0EH
   INT  21H
@@1:
   XCHG AX,CX           { Put the return value into AX }
end;

procedure ParseFile(const Src          :String;
                    var Dest           :TParsed);
assembler;
var
   Buffer          :Array [1..$25] of Byte;
asm
   cld
   push ds
   lds  si,Src
   lodsb
   cbw
   mov  bx,ax
   mov  byte ptr [si+bx],ah
   push ss
   pop  es
   lea  di,Buffer
   mov  ax,2901h
   int  21h
   push es
   pop  ds
   lea  si,[di+1]
   les  di,Dest
   mov  cx,11
   rep  movsb
   pop  ds
end;

{ TDeleteList }

constructor TDeleteList.Init;
var
   Cfg             :Text;
   S               :String;
   Op              :String[15];
   Ln, i, p        :Integer;
   NewType         :^TTypeInfo;
   NewMask         :^TMaskInfo;

procedure GetParam(Q                   :Boolean);
var
   i               :Integer;
begin
   p:=Pos(' ', S);
   Op:=S; if p>0 then Op[0]:=Chr(p-1);
   for i:=1 to Length(Op) do Op[i]:=UpCase(Op[i]);

   System.Delete(S, 1, p);
   if (S='') or (Q and ((S[1]<>'''') or (S[Length(S)]<>''''))) then
   begin
      WriteLn('Invalid parameter in '+CfgFile+', line #', Ln); Halt;
   end;
   if Q then
   begin
      System.Delete(S, 1, 1); Dec(S[0]);
   end;
end;

begin
   Inherited Init(16, 16);
   Types.Init(4, 4);

   Assign(Cfg, CfgFile); Reset(Cfg);
   Ln:=0;

   while not EOF(Cfg) do
   begin
      ReadLn(Cfg, S); Inc(Ln);

      while (S<>'') and ((S[1]=' ') or (S[1]=#9)) do System.Delete(S, 1, 1);
      while (S<>'') and ((S[Length(S)]=' ') or (S[Length(S)]=#9)) do Dec(S[0]);

      if (S='') or (S[1]=';') then Continue;

      GetParam(False);
      if Op='DEFINE' then
      begin
         GetParam(True);
         New(NewType);
         NewType^.Id:=Op; NewType^.Name:=NewStr(S);
         NewType^.Count:=0; NewType^.Size:=0;
         Types.Insert(NewType);
      end else
      begin
         NewMask:=nil;
         for i:=0 to Types.Count-1 do
         if TTypeInfo(Types.At(i)^).Id=Op then
         begin
            New(NewMask);
            NewMask^.Info:=i;
            ParseFile(S, NewMask^.Mask);
            Insert(NewMask);
            Break;
         end;
         if NewMask=nil then
         begin
            WriteLn('Invalid command in '+CfgFile+', line #', Ln); Halt;
         end;
      end;
   end;
   Close(Cfg);

   if IOResult<>0 then
   begin
      WriteLn('Unable to read '+CfgFile); Halt;
   end;
   if Count=0 then
   begin
      WriteLn('No files defined in '+CfgFile); Halt;
   end;
end;

{ TSweeper }

constructor TSweeper.Init;
begin
   Inherited Init;

   DeleteList.Init;
   TempList.Init(64, 256);
end;

procedure TSweeper.Run;
var
   Drive           :Char;
begin
   for Drive:='C' to 'Z' do
   if DriveValid(Drive) then
   begin
      WriteLn('Scanning disk ', Drive, ':');

      asm
         mov  ah,1Ch
         mov  dl,Drive
         sub  dl,'A'-1
         push ds
         int  21h
         pop  ds
         les  di,Self
         xor  ah,ah
         mul  cx
         mov  es:[di].TSweeper.ClusterSize,ax
      end;

      Inherited Run(Drive+':\');
   end;
end;

procedure TSweeper.ProcessFile;
var
   Parsed          :TParsed;
   MaskInfo        :^TMaskInfo;
   Fi, Ci          :Integer;
   Matches         :Boolean;
   NewFile         :^TFileInfo;
   TypeInfo        :^TTypeInfo;
   Remainder       :Word;
   RealSize        :LongInt;
begin
   ParseFile(Info.Name, Parsed);
   for Fi:=0 to DeleteList.Count-1 do
   begin
      MaskInfo:=DeleteList.At(Fi);
      Matches:=True;
      for Ci:=Low(TParsed) to High(TParsed) do
      if (MaskInfo^.Mask[Ci]<>'?') and (Parsed[Ci]<>MaskInfo^.Mask[Ci]) then
      begin
         Matches:=False; Break;
      end;

      if Matches then
      begin
         Remainder:=Info.Size mod ClusterSize;
         if Remainder=0
         then RealSize:=Info.Size
         else RealSize:=Info.Size - Remainder + ClusterSize;

         New(NewFile);
         NewFile^.Name:=NewStr(Dir+Info.Name); NewFile^.Info:=MaskInfo^.Info;
         TempList.Insert(NewFile);

         Inc(TempSize, RealSize);

         TypeInfo:=DeleteList.Types.At(MaskInfo^.Info);
         Inc(TypeInfo^.Count);
         Inc(TypeInfo^.Size, RealSize);

         Break;
      end;
   end;
end;

procedure TSweeper.Delete;
var
   Ti, Fi          :Integer;
   DelCount        :Integer;
   DelSize         :LongInt;
   FileInfo        :^TFileInfo;
   Plural          :String[1];
   Key             :Char;
   F               :File;
begin
   if TempList.Count=0 then
   begin
      WriteLn(#13#10'No files to delete'); Exit;
   end;

   if TempList.Count=1 then Plural:='' else Plural:='s';
   WriteLn(#13#10'Found ', TempList.Count, ' file', Plural, ' to delete, occupying ', TempSize shr 10, ' Kb');
   if TempList.Count>1 then WriteLn('Among them:');

   with DeleteList.Types do
   for Ti:=0 to Count-1 do
   with TTypeInfo(At(Ti)^) do
   if Count<>0 then
   begin
      if Count=1 then Plural:='' else Plural:='s';
      WriteLn(Count:5, ' ', Name^, Plural, ', occupying ', Size shr 10, ' Kb');
   end;
   WriteLn;

   DelCount:=0; DelSize:=0;
   with DeleteList.Types do
   for Ti:=0 to Count-1 do
   with TTypeInfo(At(Ti)^) do
   if Count<>0 then
   repeat
      if Count=1 then Plural:='' else Plural:='s';
      Write('Delete the ', Name^, Plural, ' [Yes/No/List]? ');
      asm
@@Repeat:
         xor  ax,ax
         int  16h
         and  al,not 20h
         cmp  al,'N'
         je   @@Done
         cmp  al,'Y'
         je   @@Done
         cmp  al,'L'
         jne  @@Repeat
@@Done:
         mov  Key,al
         int  29h
      end;
      WriteLn;
      case Key of
         'Y':
         begin
            Inc(DelCount, Count); Inc(DelSize, Size);

            for Fi:=0 to TempList.Count-1 do
            begin
               FileInfo:=TempList.At(Fi);
               if FileInfo^.Info=Ti then
               begin
                  Assign(F, FileInfo^.Name^); Erase(F); InOutRes:=0;
               end;
            end;
         end;
         'L':
         begin
            WriteLn;
            for Fi:=0 to TempList.Count-1 do
            begin
               FileInfo:=TempList.At(Fi);
               if FileInfo^.Info=Ti then
                  WriteLn(FileInfo^.Name^);
            end;
            WriteLn;
         end;
      end;
   until Key<>'L';

   if DelCount=0 then WriteLn(#13#10'No files deleted') else
   begin
      if DelCount=1 then Plural:='' else Plural:='s';
      WriteLn(#13#10'Deleted ', DelCount, ' file', Plural, ', occupying ', DelSize shr 10, ' Kb');
   end;
end;

var
   S               :TSweeper;

begin
   TextRec(Output).Handle:=2;
   TextRec(Output).BufSize:=1;

   WriteLn(#13#10'Disk Sweeper  Version '+VStr+'  Copyright (c) 1995 by Solar Designer \ BPC'#13#10);
   S.Init;
   S.Run;
   S.Delete;
end.

